#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>

#include <linux/pcmcia.h>
#include "cis_defs.h"
#include "pcmciad.h"

struct reg_t {
     char addr;
     char *name;
     int width;
};

struct reg_t pcic_regs[] = {
     PCIC_ID_REV, "Identification and Revision", 1,
     PCIC_STATUS, "Interface Status", 1,
     PCIC_POWER, "Power and RESETDRV control", 1,
     PCIC_INT_GEN, "Interrupt and General Control", 1,
     PCIC_STAT_CHG, "Card Status Change", 1,
     PCIC_STAT_INT, "Card Status Change Interrupt Config", 1,
     PCIC_CD_GENCTL, "Card Detect and General Control", 1,
     PCIC_GLOBAL_CTL, "Global Control", 1,
     PCIC_ADDRWINE, "Address Window Enable", 1,
     PCIC_IOCTL, "I/O Control", 1,
     PCIC_IO0_STL, "I/O Address 0 Start", 2,
     PCIC_IO0_SPL, "I/O Address 0 Stop", 2,
     PCIC_IO1_STL, "I/O Address 1 Start", 2,
     PCIC_IO1_SPL, "I/O Address 1 Stop", 2,
     PCIC_SM0_STL, "System Memory Address 0 Mapping Start", 2,
     PCIC_SM0_SPL, "System Memory Address 0 Mapping Stop", 2,
     PCIC_CM0_L, "Card Memory Offset Address 0", 2,
     PCIC_SM1_STL, "System Memory Address 1 Mapping Start", 2,
     PCIC_SM1_SPL, "System Memory Address 1 Mapping Stop", 2,
     PCIC_CM1_L, "Card Memory Offset Address 1", 2,
     PCIC_SM2_STL, "System Memory Address 2 Mapping Start", 2,
     PCIC_SM2_SPL, "System Memory Address 2 Mapping Stop", 2,
     PCIC_CM2_L, "Card Memory Offset Address 2", 2,
     PCIC_SM3_STL, "System Memory Address 3 Mapping Start", 2,
     PCIC_SM3_SPL, "System Memory Address 3 Mapping Stop", 2,
     PCIC_CM3_L, "Card Memory Offset Address 3", 2,
     PCIC_SM4_STL, "System Memory Address 4 Mapping Start", 2,
     PCIC_SM4_SPL, "System Memory Address 4 Mapping Stop", 2,
     PCIC_CM4_L, "Card Memory Offset Address 4", 2,
     0, NULL, 0,
};     

struct reg_t pcmcia_regs[] = {
     PCMCIA_COR, "Configuration Option Register", 1,
     PCMCIA_CCSR, "Card Configuration Status Register", 1,
     PCMCIA_PIR, "Pin Replacement Register", 1,
     PCMCIA_SCR, "Socket and Copy Register", 1,
     0, NULL, 0,
};

main(int argc, char **argv)
{
     struct pcic_sock pcic_socks[MAX_SOCKS], *sock;
     int i;

#ifdef USER_DRV
     pcmcia_init(0, 0);
#endif

     memset((char *) pcic_socks, 0, sizeof(pcic_socks));

     for (i = 0; i < MAX_SOCKS; i++) {
	  sock = &pcic_socks[i];
	  pcic_init_sock(i, sock);
	  
	  if (!sock->valid || !sock->cardp)
	       continue;

	  pcic_dump_regs(sock);
	  if (argc == 2)
	       pcmcia_dump_regs(sock);
     }

     return 0;
}

pcic_dump_regs(struct pcic_sock *sock)
{
     struct pcmcia_register reg;
     unsigned short val;
     int i;

     for (i = 0; pcic_regs[i].name; i++) {
	  val = 0;
	  switch (pcic_regs[i].width) {
	  case 2:
	       reg.reg = (pcic_regs[i].addr + 1) | 0x80;
	       if (ioctl(sock->fd, PCMCIA_GET_REG, &reg) < 0) {
		    perror("getting i365 register");
		    return;
	       }
	       val = reg.val << 8;
	  case 1:
	       reg.reg = pcic_regs[i].addr | 0x80;
	       if (ioctl(sock->fd, PCMCIA_GET_REG, &reg) < 0) {
		    perror("getting i365 register");
		    return;
	       }
	       val |= reg.val;
	  }
	  fprintf(stderr, "%s: %#x\n", pcic_regs[i].name, val);
     }
}

pcmcia_dump_regs(struct pcic_sock *sock)
{
     struct pcmcia_register reg;
     int i;

     reg.ofs = sock->base_addr;
     
     for (i = 0; pcmcia_regs[i].name; i++) {
	  reg.reg = pcmcia_regs[i].addr;
	  if (ioctl(sock->fd, PCMCIA_GET_REG, &reg) < 0) {
	       perror("getting PCMCIA register");
	       return;
	  }
	  printf("%s: %#x\n", pcmcia_regs[i].name, reg.val);
     }
}
