static const char *pcmciad_c_rcsid = "$Id: pcmciad.c,v 1.6 1994/04/27 00:45:27 bjaspan Exp $";

#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/errno.h>

#include <linux/pcmcia.h>
#include "pcmciad.h"

#ifdef DEBUG
#define syslog fprintf
#define LOG_NOTICE stderr
#define LOG_INFO stderr
#define LOG_ERR stderr
#define SYSLOG(args) syslog args
#else
#define SYSLOG(args)
#endif

void pcic_init_sock(int i, struct pcic_sock *sock);
void pcic_parse_card(struct pcic_sock *sock, int print_tuples);
void enable_card(struct pcic_sock *sock, struct dev_action *act,
		 struct pcmcia_cfentry *cfent);

#define IOCTL(msg, cmd, arg) \
     do { int _val; _val=arg; IOCTL_PTR(msg,cmd,&_val); } while (0)
#define IOCTL_PTR(msg, cmd, arg) \
     do { if(ioctl(sock->fd,cmd,arg)<0) perror(msg); } while (0)
     
main(int argc, char **argv)
{
     struct pcic_sock pcic_socks[MAX_SOCKS], *sock;
     extern struct dev_action *dev_actions[];
     struct dev_action *act;
     struct pcmcia_cfentry *cfent;
     int i, j, ret, print_tuples;

     print_tuples = 0;
     
     argc--; argv++;
     while (argc) {
	  if (strcmp(*argv, "-pt") == 0)
	       print_tuples++;
	  else
	       break;
	  argc--; argv++;
     }
     if (argc) {
	  fprintf(stderr, "Usage: pcmciad [-pt]\n");
	  exit(1);
     }

#ifdef USER_DRV
     pcmcia_init(0, 0);
#endif
     
     parse_configfile(stdin);

     memset((char *) pcic_socks, 0, sizeof(pcic_socks));

     for (i = 0; i < MAX_SOCKS; i++) {

	  sock = &pcic_socks[i];
	  pcic_init_sock(i, sock);
	  
	  if (sock->cardp && sock->state == EMPTY) {

	       /* card inserted */

	       pcic_parse_card(sock, print_tuples);

	       sock->state = FULL;
	       sock->devtype = lookup_devtype(sock->manufacturer,
					      sock->prod_name,
					      sock->addl_info1,
					      sock->addl_info2);
	       
	       SYSLOG((LOG_INFO, "%s card installed in slot %d\n",
		       unparse_devtype(sock->devtype), i));

	       act = dev_actions[sock->devtype];
	  next_action:
	       /* locate the next free action for devtype */
	       for (; act; act = act->next) {
		    if (!act->busy)
			 break;
	       }
	       
	       if (act == NULL) {
		    syslog(LOG_NOTICE, "all defined %s cards "
			   "already installed; ignoring\n",
			   unparse_devtype(sock->devtype));
		    continue;
	       }

	       /* look for a matching config table entry */
	       for (j = 0; j < MAX_CONFIGS; j++) {
		    cfent = &sock->configs[j];
		    if (act_matches(act, cfent))
			 break;
		    else if (sock->config_midx == cfent->idx) {
			 j = MAX_CONFIGS;
			 break; /* no "continue 2;" */
		    }
	       }
	       if (j == MAX_CONFIGS) {
		    syslog(LOG_NOTICE, "%s action not found on card; "
			   "ignoring\n",
			   unparse_devtype(sock->devtype));
		    goto next_action;
	       }
	       
	       enable_card(sock, act, cfent);
	       syslog(LOG_INFO, "initializing with \"%s\"\n", act->init);
	       ret = system(act->init);
	       if (WIFEXITED(ret)) {
		    if (WEXITSTATUS(ret)) {
			 syslog(LOG_NOTICE, "init action exited with "
				"status %d\n", WEXITSTATUS(ret));
		    }
	       } else
		    syslog(LOG_NOTICE, "init action exited on signal %d\n",
			   WTERMSIG(ret));
	  }
     }

     return 0;
}

int lookup_devtype(char *manufacturer, char *prod_name,
		   char *addl_info1, char *addl_info2)
{
     extern struct prod_map *prod_map;
     struct prod_map *pmap;
     int i;

     for (pmap = prod_map; pmap; pmap = pmap->next) {
	  if (pmap->devtype == UNKNOWN)
	       break;
	  /* XXX this should do a regexp compare */
	  else if ((!strcmp(pmap->manufacturer, ".*") || 
		    !strcmp(pmap->manufacturer, manufacturer)) &&
		   (!strcmp(pmap->prod_name, ".*") ||
		    !strcmp(pmap->prod_name, prod_name)) &&
		   (!strcmp(pmap->addl_info1, ".*") ||
		    !strcmp(pmap->addl_info1, addl_info1)) &&
		   (!strcmp(pmap->addl_info2, ".*") ||
		    !strcmp(pmap->addl_info2, addl_info2)))
	       return pmap->devtype;
     }
     return UNKNOWN;
}

int act_matches(struct dev_action *act, struct pcmcia_cfentry *cfent)
{
     static struct pcmcia_cfentry *def_cfent = NULL;

     if (cfent->defp)
	  def_cfent = cfent;

     if (cfent->irqp) {
	  if (! irq_matches(act, cfent))
	       return 0;
     } else if (def_cfent && def_cfent->irqp) {
	  if (! irq_matches(act, def_cfent))
	       return 0;
     } else if (act->cfent.irqp)
	  return 0;
     else
	  return 1;

#define INTS_MATCH(p, val, len) \
     if (cfent->p) { \
	  if (! ints_match(act->cfent.val, act->cfent.len, \
			   cfent->val, cfent->len)) \
	       return 0; \
     } else if (def_cfent && def_cfent->p) { \
	  if (! ints_match(act->cfent.val, act->cfent.len, \
			   def_cfent->val, def_cfent->len)) \
	       return 0; \
     } else if (act->cfent.p) \
	  return 0; \
     else \
	  return 1;

     INTS_MATCH(iop, io_addrs, ios);
     INTS_MATCH(iop, io_lens, ios);
     INTS_MATCH(memp, mem_caddrs, mems);
     INTS_MATCH(memp, mem_haddrs, mems);
     INTS_MATCH(memp, mem_lens, mems);
#undef INTS_MATCH

     return 1;
}

int irq_matches(struct dev_action *act, struct pcmcia_cfentry *cfent)
{
     return cfent->irq ? cfent->irq == act->cfent.irq :
	  (cfent->irq_mask & (1 << act->cfent.irq));
}

int ints_match(int *addrs1, int num1, int *addrs2, int num2)
{
     if (num1 == num2 && !memcmp(addrs1, addrs2, sizeof(int)*num1))
	  return 1;
     return 0;
}

void enable_card(struct pcic_sock *sock, struct dev_action *act,
		 struct pcmcia_cfentry *cfent)
{
     struct pcmcia_io_map io_map;
     struct pcmcia_mem_map mem_map;
     struct pcmcia_register reg;
     int i, val;

     /* Set card type = I/O */
     IOCTL_PTR("getting flags", PCMCIA_GET_FLAGS, &val);
     val |= PC_IOCARD;
     IOCTL("setting type = I/O", PCMCIA_SET_FLAGS, val);

     /* Unmap all I/O regions... XXX should be ioctl */
     memset((char *) &io_map, 0, sizeof(io_map));
     for (i = 0; i < 2; i++) {
	  io_map.map = i;
	  IOCTL_PTR("unmapping I/O", PCMCIA_SET_IO_MAP, &io_map);
     }

     /* Map I/O regions */
     for (i = 0; i < act->cfent.ios; i++) {
	  io_map.map = i;
	  io_map.start = act->cfent.io_addrs[i];
	  io_map.stop =  act->cfent.io_addrs[i] + act->cfent.io_lens[i] - 1;
	  /* XXX does CIS determine MAP_CS16? */
	  io_map.flags = MAP_ACTIVE |
	       (sock->devtype == ETHERNET ? (i == 1 ? MAP_CS16 : 0)
		: MAP_CS16);
	  IOCTL_PTR("mapping I/O region", PCMCIA_SET_IO_MAP, &io_map);
     }

     /* Unmap all memory regions... XXX should be ioctl */
     memset((char *) &mem_map, 0, sizeof(mem_map));
     for (i = 0; i < 4; i++) {
	  mem_map.map = i;
	  IOCTL_PTR("unmapping memory", PCMCIA_SET_MEM_MAP, &mem_map);
     }
     
     /* Map memory regions */
     for (i = 0; i < act->cfent.mems; i++) {
	  mem_map.map = i;
	  mem_map.flags = MAP_ACTIVE | MAP_ATTRIB;
	  mem_map.extra_ws = 0;
	  mem_map.sys_start = act->cfent.mem_haddrs[i];
	  mem_map.sys_stop = act->cfent.mem_haddrs[i] +
	       act->cfent.mem_lens[i] - 1;
	  
	  mem_map.card_start = 0; /* XXX act->cfent.mem_caddrs[i]; */
	  IOCTL_PTR("mapping memory", PCMCIA_SET_MEM_MAP, &mem_map);
     }

     IOCTL("setting irq", PCMCIA_SET_IRQ, act->cfent.irq);

     /* Write config index to COR */
     /* XXX PCMCIA_MEMIO is *not* documented in my PCMCIA specs */
     reg.ofs = sock->base_addr;
     reg.reg = PCMCIA_COR;
     reg.val = cfent->idx | PCMCIA_MEMIO | PCMCIA_LVLREQ;
     IOCTL_PTR("setting COR", PCMCIA_SET_REG, &reg);
}


